import { EntityComponentTypes, EquipmentSlot, ItemStack, world } from "@minecraft/server";
import { Dimension, herobrineData, herobrineEvents, herobrineSettings } from "./index.js";

export const updateHerobrineData = () => { writeDynamicData("herobrineData", herobrineData) }
export const updateDynamicSettings = () => { writeDynamicData("herobrineSettings", herobrineSettings) }
export const updateDynamicEvents = () => { writeDynamicData("herobrineEvents", herobrineEvents) }

export const findIndexPerString = (list, input = "test") => list.findIndex(obj => { return input.toLocaleLowerCase() == obj.name.toLocaleLowerCase() })

export function equipRandomItemsToZombie(zombie) {
    const swordsList = ["minecraft:iron_sword", "minecraft:stone_sword", "minecraft:golden_sword", "minecraft:wooden_sword"];

    const helmetList = ["minecraft:iron_helmet", "minecraft:golden_helmet", "minecraft:chainmail_helmet"];
    const chestplateList = ["minecraft:iron_chestplate", "minecraft:golden_chestplate", "minecraft:chainmail_chestplate"];
    const leggingsList = ["minecraft:iron_leggings", "minecraft:golden_leggings", "minecraft:chainmail_leggings"];
    const bootsList = ["minecraft:iron_boots", "minecraft:golden_boots", "minecraft:chainmail_boots"];

    zombie.runCommand(`replaceitem entity @s slot.weapon.mainhand 0 ${swordsList[randomNumber(0, swordsList.length - 1)]}`)

    zombie.runCommand(`replaceitem entity @s slot.armor.head 0 ${helmetList[randomNumber(0, helmetList.length - 1)]}`)
    zombie.runCommand(`replaceitem entity @s slot.armor.chest 0 ${chestplateList[randomNumber(0, chestplateList.length - 1)]}`)
    zombie.runCommand(`replaceitem entity @s slot.armor.legs 0 ${leggingsList[randomNumber(0, leggingsList.length - 1)]}`)
    zombie.runCommand(`replaceitem entity @s slot.armor.feet 0 ${bootsList[randomNumber(0, bootsList.length - 1)]}`)

    zombie.addEffect("fire_resistance", 20 * 60, { showParticles: false })
}

export function calculateDistance(vetor1, vetor2) {
    const dx = vetor2.x - vetor1.x;
    const dy = vetor2.y - vetor1.y;
    const dz = vetor2.z - vetor1.z;
    return Math.sqrt(dx * dx + dy * dy + dz * dz).toFixed(3);
}

export function herobrineVFX(location, id = "minecraft:ice_evaporation_emitter") {
    const step = 0.1;
    const minHeight = -0.6;
    const maxHeight = 0.80;
    const radius = 0.65;
    const particles = 5;

    for (let height = minHeight; height < maxHeight; height += step) {
        VFX(location, id, radius, particles, height);
    }
}

export function VFX(location, particleID, radius, maxParticles = 100, yFIX = 0) {
    const clamp = (value, min, max) => Math.max(min, Math.min(value, max));
    const { x: centerX, y: centerY, z: centerZ } = location;
    const particleCount = Math.min(maxParticles, Math.floor(360 / radius));

    for (let i = 0; i < particleCount; i++) {
        const theta = Math.random() * 2 * Math.PI;
        const phi = Math.acos(2 * Math.random() - 1);

        const distance = radius * Math.random() * 0.5;

        const particleX = centerX + distance * Math.sin(phi) * Math.cos(theta) - 0.5;
        const particleY = clamp(centerY + distance * Math.cos(phi) + yFIX, -64, 320);
        const particleZ = centerZ + distance * Math.sin(phi) * Math.sin(theta) - 0.5;

        Dimension.spawnParticle(particleID, { x: particleX, y: particleY, z: particleZ });
    }
}

export function randomNumber(min, max, round = true) {
    if (min > max) return null;

    const value = Math.random() * (max - min) + min;
    return round ? Math.round(value) : value;
}

export function fetchDynamicData(name) {
    try {
        return JSON.parse(world.getDynamicProperty(name))
    } catch {
        console.warn("Erro na conversão, pode ser que esse Dynamic não exista.", name)
        return null
    }
}

export function writeDynamicData(name, input = [{ none: true }]) {
    const inputFormat = JSON.stringify(input)
    world.setDynamicProperty(name, inputFormat)
}

export function updateDynamicData(dynamicName, structureConfig = null) {
    try {

        const dynamicData = fetchDynamicData(dynamicName);

        if (dynamicData == null || dynamicData == undefined) {
            writeDynamicData(dynamicName, structureConfig);
        }

        return dynamicData

    } catch (error) {
        writeDynamicData(dynamicName, structureConfig);
        return fetchDynamicData(dynamicName);
    }
}

export async function spawnEntityInRegion(entityTypeId, centerLocation, radius = 4, searchHeight = 1, minSpawnDistance = 5) {
    try {
        const { x: centerX, y: centerY, z: centerZ } = centerLocation;

        // Assume que BlockChecker está instanciado corretamente em outro lugar ou instancie aqui se necessário
        const blockChecker = new BlockChecker(); // Certifique-se que isso está acessível

        const xMin = Math.floor(centerX - radius);
        const xMax = Math.floor(centerX + radius);
        const zMin = Math.floor(centerZ - radius);
        const zMax = Math.floor(centerZ + radius);
        const yBase = Math.floor(centerY);
        const yMin = Math.max(-64 + 1, yBase - searchHeight);
        const yMax = Math.min(319 - 1, yBase + searchHeight);

        let chosenLocation = null;
        let chosenFloorBlock = null;
        let validLocationCount = 0;

        for (let potentialY = yMin; potentialY <= yMax; potentialY++) {
            if (potentialY <= -64 || potentialY >= 319) continue;

            for (let potentialX = xMin; potentialX <= xMax; potentialX++) {
                for (let potentialZ = zMin; potentialZ <= zMax; potentialZ++) {

                    const distanceX = Math.abs(potentialX - centerX);
                    const distanceZ = Math.abs(potentialZ - centerZ);

                    if (distanceX < minSpawnDistance && distanceZ < minSpawnDistance) {
                        continue;
                    }

                    let blockCurrent, blockAbove, blockFloor;
                    try {
                        blockCurrent = Dimension.getBlock({ x: potentialX, y: potentialY, z: potentialZ });
                        blockAbove = Dimension.getBlock({ x: potentialX, y: potentialY + 1, z: potentialZ });
                        blockFloor = Dimension.getBlock({ x: potentialX, y: potentialY - 1, z: potentialZ });
                    } catch (e) {
                        continue;
                    }

                    if (!blockCurrent || !blockAbove || !blockFloor) {
                        continue;
                    }

                    const isSpaceAvailable = blockCurrent.isAir && blockAbove.isAir;
                    const isFloorSolid = !blockFloor.isAir && !blockFloor.isLiquid;
                    const isFloorSpawnable = !blockChecker.isSpawnDenied(blockFloor.typeId); // Usa blockChecker

                    if (isSpaceAvailable && isFloorSolid && isFloorSpawnable) {
                        validLocationCount++;
                        if (Math.random() < 1 / validLocationCount) {
                            chosenLocation = { x: potentialX, y: potentialY, z: potentialZ };
                            chosenFloorBlock = blockFloor;
                        }
                    }
                }
            }
        }

        if (chosenLocation && chosenFloorBlock) {
            try {
                // 1. Pega propriedades do bloco chão
                const floorProperties = blockChecker.getBlockProperties(chosenFloorBlock.typeId);
                // 2. Pega altura do bloco chão (padrão 1 se não especificado)
                const floorHeight = floorProperties.height ?? 1; // Usa ?? 1 como fallback se height for undefined/null
                // 3. Y base do bloco chão
                const floorSurfaceY = chosenLocation.y;
                // 4. Calcula o Y da superfície somando a altura do bloco chão
                const spawnY = floorSurfaceY - floorHeight;

                // 5. Define a posição final de spawn com um pequeno ajuste vertical
                const smallVerticalOffset = 0.01; // Pequeno deslocamento para evitar afundar
                const spawnPos = {
                    x: chosenLocation.x + 0.5, // Centraliza no bloco X
                    y: spawnY + smallVerticalOffset, // Posição Y na superfície + pequeno deslocamento
                    z: chosenLocation.z + 0.5  // Centraliza no bloco Z
                };

                // Logs para depuração (opcional)
                // console.log(`[Spawn] Tentando spawnar ${entityTypeId} em ${JSON.stringify(spawnPos)}.`);
                // console.log(`[Spawn] Bloco chão: ${chosenFloorBlock.typeId} em Y=${floorSurfaceY}, Altura: ${floorHeight}, Spawn Y Base: ${spawnY}`);


                const spawnedEntity = Dimension.spawnEntity(entityTypeId, spawnPos);
                return spawnedEntity;

            } catch (spawnError) {
                console.error(`Falha ao spawnar entidade ${entityTypeId} na posição ${JSON.stringify(chosenLocation)} (calculada: ${JSON.stringify(spawnPos)}):`, spawnError);
                return null;
            }
        } else {
            // console.log(`Nenhuma posição válida encontrada para ${entityTypeId} na região.`);
            return null;
        }

    } catch (error) {
        console.error(`Erro geral na função spawnEntityInRegion para ${entityTypeId}:`, error);
        if (error.stack) {
            console.error(error.stack);
        }
        return null;
    }
}

export class BlockManager {

    replaceBlock(vector3, blockInLoc, replacementBlock, keepPermutation) {
        if (keepPermutation) {
            const permutationData = blockInLoc.permutation.getState("torch_facing_direction");
            Dimension.setBlockType(vector3, replacementBlock);
            Dimension.setBlockPermutation(
                vector3,
                blockInLoc.permutation.withState("torch_facing_direction", permutationData)
            );
        } else {
            Dimension.setBlockType(vector3, replacementBlock);
        }
    }

    squareGetBlocks({
        locationStart,
        allSides = false,
        blockTargetID = "none",
        replacementBlock = "minecraft:redstone_torch",
        keepPermutation = true,
        rad = 30,
        height = 5
    } = {}) {
        try {
            const { x: originX, y: originY, z: originZ } = locationStart;

            const [minX, minY, minZ] = [
                Math.round(originX - rad),
                Math.round(allSides ? originY - height : originY),
                Math.round(originZ - rad)
            ];
            const [maxX, maxY, maxZ] = [
                Math.round(originX + rad),
                Math.round(originY + height),
                Math.round(originZ + rad)
            ];

            for (let x = minX; x <= maxX; x++) {
                for (let y = minY; y <= maxY; y++) {
                    for (let z = minZ; z <= maxZ; z++) {
                        const vector3 = { x: x, y: y, z: z };
                        const blockInLoc = Dimension.getBlock(vector3);

                        if (blockTargetID !== "none" && blockInLoc.typeId === blockTargetID && blockInLoc.typeId !== "minecraft:air") {
                            this.replaceBlock(vector3, blockInLoc, replacementBlock, keepPermutation);
                        }
                    }
                }
            }
        } catch { }
    }
};


export class BlockChecker {

    isSpawnDenied(blockId, blockDataList = blockData_list) {
        let normalizedBlockId = blockId;
        if (!normalizedBlockId.startsWith("minecraft:")) {
            normalizedBlockId = "minecraft:" + normalizedBlockId;
        }

        if (blockDataList.notAllowed.includes(normalizedBlockId)) {
            return true;
        }

        if (blockDataList.blockTypeid.hasOwnProperty(normalizedBlockId)) {
            const properties = blockDataList.blockTypeid[normalizedBlockId];
            return properties.canSpawnOn === false;
        }

        for (const suffix in blockDataList.suffixProperties) {
            if (blockDataList.suffixProperties.hasOwnProperty(suffix)) {
                if (normalizedBlockId.endsWith(suffix)) {
                    const properties = blockDataList.suffixProperties[suffix];
                    return properties.canSpawnOn === false;
                }
            }
        }

        return false;
    }

    getBlockProperties(blockId, blockDataList = blockData_list) {
        let normalizedBlockId = blockId;
        if (!normalizedBlockId.startsWith("minecraft:")) {
            normalizedBlockId = "minecraft:" + normalizedBlockId;
        }

        if (blockDataList.notAllowed.includes(normalizedBlockId)) {
            let height = 1;
            if (blockDataList.blockTypeid.hasOwnProperty(normalizedBlockId)) {
                height = blockDataList.blockTypeid[normalizedBlockId].height;
            } else {
                for (const suffix in blockDataList.suffixProperties) {
                    if (blockDataList.suffixProperties.hasOwnProperty(suffix) && normalizedBlockId.endsWith(suffix)) {
                        height = blockDataList.suffixProperties[suffix].height;
                        break;
                    }
                }
            }
            return { height: height, canSpawnOn: false, source: 'notAllowed' };
        }

        if (blockDataList.blockTypeid.hasOwnProperty(normalizedBlockId)) {
            return { ...blockDataList.blockTypeid[normalizedBlockId], source: 'blockTypeid' };
        }

        for (const suffix in blockDataList.suffixProperties) {
            if (blockDataList.suffixProperties.hasOwnProperty(suffix)) {
                if (normalizedBlockId.endsWith(suffix)) {
                    return { ...blockDataList.suffixProperties[suffix], source: 'suffixProperties' };
                }
            }
        }

        return { height: 0, canSpawnOn: true, source: 'default' };
    }
}

const blockData_list = {
    // Blocos que não devem permitir spawn (mesmo que “permitidos” para retornar, não serão usados para spawnar)
    notAllowed: [
        "minecraft:bamboo",
        "minecraft:fire",
        "minecraft:water",
        "minecraft:lava",
        "minecraft:tall_grass",
        "minecraft:rose_bush",
        "minecraft:peony",
        "minecraft:lilac",
        "minecraft:sunflower"
    ],
    // Blocos com propriedades definidas por nome completo
    blockTypeid: {
        // Plantas altas (2 blocos) – spawn desabilitado
        "minecraft:rose_bush": { height: 2, canSpawnOn: false },
        "minecraft:lilac": { height: 2, canSpawnOn: false },
        "minecraft:peony": { height: 2, canSpawnOn: false },
        "minecraft:sunflower": { height: 2, canSpawnOn: false },
        "minecraft:tall_grass": { height: 2, canSpawnOn: false },

        // Plantas pequenas (1 bloco)
        "minecraft:wildflowers": { height: 0, canSpawnOn: true },
        "minecraft:dandelion": { height: 1, canSpawnOn: true },
        "minecraft:poppy": { height: 1, canSpawnOn: true },
        "minecraft:blue_orchid": { height: 1, canSpawnOn: true },
        "minecraft:allium": { height: 1, canSpawnOn: true },
        "minecraft:azure_bluet": { height: 1, canSpawnOn: true },
        "minecraft:red_tulip": { height: 1, canSpawnOn: true },
        "minecraft:orange_tulip": { height: 1, canSpawnOn: true },
        "minecraft:white_tulip": { height: 1, canSpawnOn: true },
        "minecraft:pink_tulip": { height: 1, canSpawnOn: true },
        "minecraft:oxeye_daisy": { height: 1, canSpawnOn: true },
        "minecraft:cornflower": { height: 1, canSpawnOn: true },
        "minecraft:lily_of_the_valley": { height: 1, canSpawnOn: true },
        "minecraft:fern": { height: 1, canSpawnOn: true },
        "minecraft:dead_bush": { height: 1, canSpawnOn: true },

        // Blocos especiais
        "minecraft:snow_layer": { height: 0.125, canSpawnOn: true },
        "minecraft:leaf_litter": { height: 0, canSpawnOn: true },
        "minecraft:torch": { height: 1, canSpawnOn: false },
        "minecraft:lever": { height: 1, canSpawnOn: false },

        "minecraft:daylight_detector": { height: 1, canSpawnOn: true },
        "minecraft:daylight_detector_inverted": { height: 1, canSpawnOn: true },
        "minecraft:pointed_dripstone": { height: 1, canSpawnOn: true },
        "minecraft:short_grass": { height: 0, canSpawnOn: true },
        "minecraft:heavy_core": { height: 1, canSpawnOn: true },
        "minecraft:bell": { height: 1, canSpawnOn: true },
        "minecraft:tripwire_hook": { height: 1, canSpawnOn: true }
    },
    // Propriedades para blocos que são identificados por sufixo
    suffixProperties: {
        "_slab": { height: 0.5, canSpawnOn: true },
        "_double_slab": { height: 1, canSpawnOn: true },
        "_fence": { height: 1.5, canSpawnOn: false },
        "_wall": { height: 1.5, canSpawnOn: false },
        "_stairs": { height: 1, canSpawnOn: true },
        "_carpet": { height: 0.99, canSpawnOn: true },
        "_pressure_plate": { height: 0.0625, canSpawnOn: true },
        "_fence_gate": { height: 1, canSpawnOn: true },
        "_button": { height: 1, canSpawnOn: true },
        "_wire": { height: 1, canSpawnOn: true },
        "_repeater": { height: 1, canSpawnOn: true },
        "_comparator": { height: 1, canSpawnOn: true },
        "_banner": { height: 1, canSpawnOn: true },
        "_layer": { height: 1, canSpawnOn: true },
        "_sign": { height: 0, canSpawnOn: true },
        "_leaves": { height: 1, canSpawnOn: false },
        "_rod": { height: 1, canSpawnOn: true }
    }
};